import os

from .check import check
from .report import report
from .users import add_user, list_users, remove_user, set_ads_token, display_ads_token

# Make sure data directory exists.
# This is where the user database and cite tracker information goes.
data_dir = os.path.join(os.path.dirname(__file__), "data")
if not os.path.isdir(data_dir):
    os.makedirs(data_dir)

# User database.
USER_DATABASE = os.path.join(data_dir, "users.toml")

# Database that tracks cites.
def get_user_database_path(user_id) -> str:
    user_database = os.path.join(data_dir, f"{user_id}_database.toml")
    return user_database



from urllib.parse import urlencode
import requests

from datetime import datetime

class _ADSPaper:

    def __init__(self, paper_info):
        """
        Class to store information for a single ADS paper.

        Parameters
        ----------
        paper_info : dict
            Data returned from the query for this paper
        """

        # Insert query data into this object.
        for att in paper_info.keys():
            if att == "title":
                setattr(self, att, paper_info[att][0])
            else:
                setattr(self, att, paper_info[att])

    def get_years_since_publication(self) -> float:
        """
        Return the number of years since publication for this paper.

        Requires to of had "pubdate" in the original query.

        Returns
        -------
        - : float
            Total number of years since publication
        """

        # We need to have had pubdate in our query.
        if hasattr(self, "pubdate"):

            # Convert pubdate string to datetime.
            pubyear = int(self.pubdate.split('-')[0])
            pubmonth = int(self.pubdate.split('-')[1])
            if pubmonth == 0:
                pubmonth += 1
            dt = f"{pubyear}-{pubmonth}"

            # Compute time difference from now.
            diff = datetime.now() - datetime.strptime(dt, "%Y-%m")

            # Convert to years.
            return diff.total_seconds() / 31536000
        else:
            return None

    @property
    def citations_per_year(self) -> float:
        """ 
        How many citations has this paper has per year?

        Needs to have had "pubdate" and "citation_count" in the original query.
        """

        pubyears = self.get_years_since_publication()
        if pubyears is None or not hasattr(self, "citation_count"):
            return None
        else:
            return self.citation_count / pubyears

    @property
    def link(self) -> str:
        """ Return string hyperlink to ADS page for this paper """

        # Need bibcode in original query.
        if not hasattr(self, "bibcode"):
            return None
        
        label = self.bibcode
        uri = f"https://ui.adsabs.harvard.edu/abs/{self.bibcode}/abstract"
        parameters = ''

        # OSC 8 ; params ; URI ST <name> OSC 8 ;; ST 
        escape_mask = '\033]8;{};{}\033\\{}\033]8;;\033\\'

        return escape_mask.format(parameters, uri, label)

class _ADSQuery:
    def __init__(self, q, fl, rows, request_data):
        """
        Stores the result of an individual query.

        Not designed to be called standalone, should be called from within the
        ADSQueryWrapper object.

        Parameters
        ----------
        q : string
            The query sent to ADS
        fl : string
            Properties to return from query
        rows : int
            Max number of rows to return
        request_data : requests "get" object
            The data returned from the query from the requests lib

        Attributes
        ----------
        q : string
            See parameters
        fl : string
            See parameters
        rows : int
            See parameters
        query_time : float
            Execution time of query
        query_status: int
            Query return status
        num_found : int
            Papers found as result from query
        papers : list of dicts
            Each paper's info stored in a dict
        """

        # Store the query info.
        self.q = q
        self.fl = fl
        self.rows = rows

        # Parse out the requests information for this query.
        self.parse(request_data)

    def parse(self, request_data):

        # Convert to JSON format.
        request_data = request_data.json()

        # Store some information about the query execution.
        rheader = request_data["responseHeader"]
        self.query_time = rheader["QTime"]
        self.query_status = rheader["status"]
        assert self.query_status == 0

        # Store the result.
        self.num_found = request_data["response"]["numFound"]

        # Case where max_rows wasn't big enough.
        if self.num_found > len(request_data["response"]["docs"]):
            print(f"Warning: Query {self.q} returns over max rows,"
                  f"({self.num_found} > {self.rows})",
                  "not all papers will be in the list")

        self.papers = []
        for i in range(len(request_data["response"]["docs"])):
            self.papers.append(_ADSPaper(request_data["response"]["docs"][i]))

    def get_all(self, what):
        data = []

        for i in range(len(self.papers)):
            data.append(getattr(self.papers[i], what))

        return data


class ADSQueryWrapper:
    def __init__(self, ads_token):
        """
        Class that wraps calls to the ADS API to make easy queries.

        Each method returns a _ADSQuery object, which contains information
        about the query, the reponse from the ADS API, and the result of the query. The
        result of the query is stored in a "paper" list in the _ASDQuery object, for
        which each entry is a list containing information about the papers in the query
        result. 

        Methods
        -------
        get(query_string, return_columns, max_return_rows)
            Perform a generic query to the ADS API
        citations(bibcode)
            Query the ADS API to return all cites to a given paper
        references(bibcode)
            Query the ADS API to return all references within a given paper

        """

        # ADS API token.
        self.token = ads_token

        # Log how many ADS API calls this object has used in this session.
        self.ads_api_calls = 0

        # Log how many ADS API calls remaining on our token today.
        self.ads_api_calls_remaining = None

    def __del__(self):
        """
        On program end, report how many ADS API calls were used during this
        ADSQuery object instance. We are interested in this because the number
        of API calls to ADS has a daily limit of 5000.
        """
        if self.ads_api_calls > 0:
            print(
                f"Used {self.ads_api_calls} ADS API calls in this instance,",
                f"{self.ads_api_calls_remaining} more calls can be used today",
                f"on this token.",
            )

    def _encode_string(self, query):
        """
        Encode query dict into a string.

        Needs to include "q", the query, and "fl" the fields you want returned.

        Paramaters
        ----------
        query : dict

        Example
        -------
        query = {"q": "author:McAlpine,Stuart", "fl": "citation_count", "rows": 20}
        """

        return urlencode(query)

    def get(self, q, fl, rows=20, max_attempts=3):
        """
        Perform generic query using the ADS API.

        Parameters
        ----------
        q : string
            The query sent to ADS
        fl : string
            Properties to return from query
        rows : int (optional)
            Max number of rows to return from the query
        max_attempts : int
            How many times do we try before we give up?

        Returns
        -------
        - : _ADSQuery object
            Object stores all information about query and the result
        """

        # Build query dict.
        query = {"q": q, "fl": fl, "rows": rows}

        # Convert query dict to string.
        q = self._encode_string(query)
        url = f"https://api.adsabs.harvard.edu/v1/search/query?{q}"

        # Need authorization token in header.
        headers = {"Authorization": f"Bearer:{self.token}"}

        # Make get request.
        for i in range(max_attempts):
            resp = requests.get(url, headers=headers)
            self.ads_api_calls += 1

            # Check status code.
            if resp.status_code != 200:
                print(f"Attempt {i+1} recieved status code ",
                      f"{resp.status_code} from ADS, trying again...")
                continue
            else:
                break

        # Case where we never got a good reponse from ADS
        if resp.status_code != 200:
            print(f"Recieved too many bad status codes...")
            return None

        # Look at the header to see how many queries we have left.
        self.ads_api_calls_remaining = resp.headers["X-RateLimit-Remaining"]

        return _ADSQuery(q, fl, rows, resp)

    def citations(self, bibcode, fl="title,bibcode,author,citation_count", rows=2000):
        """
        Query what papers cite a paper of a given bibcode.

        Parameters
        ----------
        bibcode : string
            The bibcode of the paper we want to know who cites
        fl : string (optional)
            Properties to return from query
        rows : int (optional)
            Max number of rows to return

        Returns
        -------
        - : _ADSQuery object
            Object stores all information about query and the result
        """

        # Make sure bibcode is a string.
        assert type(bibcode) == str

        q = f"citations(bibcode:{bibcode})"

        return self.get(q, fl, rows=rows)

    def references(self, bibcode, fl="title,bibcode,author,citation_count", rows=2000):
        """
        Query what references a paper contains.

        Parameters
        ----------
        bibcode : string
            The bibcode of the paper we want to know who cites
        fl : string (optional)
            Properties to return from query
        rows : int (optional)
            Max number of rows to return

        Returns
        -------
        - : _ADSQuery object
            Object stores all information about query and the result
        """

        # Make sure bibcode is a string.
        assert type(bibcode) == str

        q = f"references(bibcode:{bibcode})"

        return self.get(q, fl, rows=rows)



import argparse
import myads.cite_tracker as cite_tracker


def main():

    # Command line options.
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(title="subcommand", dest="subcommand")

    # Primary options
    user_parser = subparsers.add_parser(
        "author", help="Add/remove/list tracked authors in the database"
    )
    token_parser = subparsers.add_parser("token", help="Add/update ADS API token")
    report_parser = subparsers.add_parser(
        "report", help="Report current citation statistics for tracked authors"
    )
    check_parser = subparsers.add_parser(
        "check", help="Check for any new cites for tracked authors"
    )

    # User options
    user_subparser = user_parser.add_subparsers(
        title="user_subparser", dest="user_subparser"
    )
    user_subparser.add_parser("add", help="Add a new author to be tracked")
    user_subparser.add_parser("list", help="List current tracked authors")
    tmp = user_subparser.add_parser("remove", help="Remove an existing tracked author")
    tmp.add_argument("user_id", help="User ID of tracked author to remove")

    # ADS API token options
    token_subparser = token_parser.add_subparsers(
        title="token_subparser", dest="token_subparser"
    )
    tmp = token_subparser.add_parser("add", help="Add/update ADS API token")
    tmp.add_argument("ads_token", help="ADS token to add")
    token_subparser.add_parser("display", help="Display current ADS token")

    # Check options
    check_parser.add_argument(
        "--verbose", help="True for more output", action="store_true"
    )

    args = parser.parse_args()

    if args.subcommand == "user":

        # Add a user to the database
        if args.user_subparser == "add":
            cite_tracker.add_user()

        # Remove a user from the database
        elif args.user_subparser == "remove":
            cite_tracker.remove_user(args.user_id)

        # Print user list.
        elif args.user_subparser == "list":
            cite_tracker.list_users()

    elif args.subcommand == "token":

        # Set the ADS API token.
        if args.token_subparser == "add":
            cite_tracker.set_ads_token(args.ads_token)

        # Display the current ADS API token.
        if args.token_subparser == "display":
            cite_tracker.display_ads_token()

    # Report users current citation statistics
    elif args.subcommand == "report":
        cite_tracker.report()

    # Check if any new cites have been made to the user since last call
    elif args.subcommand == "check":
        cite_tracker.check(args.verbose)



import os

import myads.cite_tracker as cite_tracker
import toml


def _create_new_profile() -> dict:
    """ Get new user information. """

    first_name = input("Enter user first name: ")
    last_name = input("Enter user surname: ")
    orcid = input("Enter user orcid (optional): ")

    data = {
        "first_name": first_name,
        "last_name": last_name,
        "orcid": orcid,
    }

    return data


def _load_user_database() -> dict:
    """ Load the user database. """

    return toml.load(cite_tracker.USER_DATABASE)


def _save_user_database(data):
    """ Save the user database. """

    with open(cite_tracker.USER_DATABASE, "w") as f:
        toml.dump(data, f)


def list_users():
    """ Print all current users in the database. """

    data = _load_user_database()

    for att in data.keys():
        if "user" in att:
            print(
                f"{att}: {data[att]['first_name']} {data[att]['last_name']}",
                f"(ORCID={data[att]['orcid']})",
            )

def display_ads_token():
    """ Print all ADS API tokens in the database. """

    data = _load_user_database()

    if "ads_token" in data["metadata"].keys():
        print(f"ADS token: {data['metadata']['ads_token']}")
    else:
        print("No ADS token added yet...")


def add_user():
    """ Adds new user to the database. """

    # Does the user database exist?
    database_exists = os.path.isfile(cite_tracker.USER_DATABASE)

    # Get new profile information.
    new_user_data = _create_new_profile()

    if database_exists:
        # Add new entry.
        data = _load_user_database()
        data["metadata"]["uid_count"] += 1
        data[f"user{data['metadata']['uid_count']}"] = new_user_data
    else:
        # First entry.
        data = {"user1": new_user_data, "metadata": {"uid_count": 1}}

    # Save database-
    _save_user_database(data)


def remove_user(removeid):
    """ Remove a user from the database. """

    # Does the user database exist?
    database_exists = os.path.isfile(cite_tracker.USER_DATABASE)
    assert database_exists, "No user database yet!"

    # Load the user database.
    data = _load_user_database()

    # Remove user
    if f"user{removeid}" in data.keys():
        print(f"Removing user{removeid}...")

        # Remove cite tracker database.
        user_database = cite_tracker.get_user_database_path(f"user{removeid}")
        if os.path.isfile(user_database):
            os.remove(user_database)
            print(f"Deleted {user_database}")

        # Remove from users list.
        del data[f"user{removeid}"]
    else:
        print(
            f"No user {removeid} in database, try 'myads user list' to see current users"
        )

    # Save database-
    _save_user_database(data)


def set_ads_token(ads_token):
    """ Store the ADS API token """

    # Need the database first.
    database_exists = os.path.isfile(cite_tracker.USER_DATABASE)
    if not database_exists:
        print("No user database yet, first run --add_user")
        return

    # Load user database.
    data = _load_user_database()

    # Add the new API token.
    data["metadata"]["ads_token"] = ads_token
    print(f"Updated ADS API token to {ads_token}")

    # Save the database.
    _save_user_database(data)



import os
import toml

# Working directory of ADS package (where database will be stored).
_wd = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))

# See if database folder exists, if not make it.
if not os.path.isdir(os.path.join(_wd, "database")):
    os.mkdir(os.path.join(_wd, "database"))

# Load the config file with users ADS info.
if not os.path.isfile(os.path.join(_wd, "myinfo.toml")):
    config = None
else:
    myinfo = toml.load(os.path.join(_wd, "myinfo.toml"))
    
    config = {
        "_DATABASE_FILE": os.path.join(_wd, "database", "mydatabase.toml"),
        "_FIRST_NAME": myinfo["info"]["first_name"],
        "_LAST_NAME": myinfo["info"]["last_name"],
        "_ADS_TOKEN": myinfo["info"]["ads_token"],
    }

    del myinfo



import os

import myads.cite_tracker as cite_tracker
import toml
from myads.query import ADSQueryWrapper
from tabulate import tabulate


def _refresh_database(data, query) -> dict:
    """
    Get the most up-to-date cites to the users' papers.

    Parameters
    ----------
    data : 
    query : ADSQueryWrapper object

    Returns
    -------
    database : dict
        Details the up-to-date cites of the users' papers
    """

    database = {}

    for paper in data.papers:
        tmp = query.citations(paper.bibcode)
        database[paper.bibcode] = {}
        database[paper.bibcode]["title"] = paper.title
        database[paper.bibcode]["citations"] = tmp.get_all("bibcode")

    return database


def _print_new_cites(database, new_cite_papers):
    """
    Print any new citations to our papers since last call.

    Parameters
    ----------
    database : dict
        Current papers that cited our papers
    new_cite_papers : dict
        New papers added to database (the ones we are printing here)
    """

    # Colours for the terminal.
    BOLD = "\033[1m"
    OKCYAN = "\033[96m"
    ENDC = "\033[0m"

    # Loop over each of our papers.
    for bibcode in new_cite_papers.keys():

        # Do we have any new cites?
        new = new_cite_papers[bibcode]

        if len(new) > 0:

            # Print new cites.
            print(
                "\n",
                f"{BOLD}{OKCYAN}{len(new)} new cite(s) for "
                f"{database[bibcode]['title']}{ENDC}",
            )
            table = []
            for paper in new:
                table.append(
                    [paper.title, paper.author, paper.date[:10], paper.link,]
                )

            print(
                tabulate(
                    table,
                    tablefmt="grid",
                    maxcolwidths=[40, 40, None, 20],
                    headers=["Title", "Authors", "Date", "Bibcode"],
                )
            )


def _save_database(path, data):
    """
    Save the users cite database.

    Parameters
    ----------
    path : str
        Path to user database file
    data : dict
        Users' cite information
    """

    with open(path, "w") as f:
        toml.dump(data, f)


def check(verbose):
    """
    Check against each users' personal database to see if there are any new
    cites to their papers since the last call.

    Parameters
    ----------
    verbose : bool  
        True for more output
    """

    # Load the user database.
    users = toml.load(cite_tracker.USER_DATABASE)
    assert (
        "ads_token" in users["metadata"].keys()
    ), "No ADS API token has been added yet, run 'myads --set_ads_token <TOKEN>'"

    # Query object.
    query = ADSQueryWrapper(users["metadata"]["ads_token"])

    # Loop over each user in the database.
    for att in users.keys():
        if "user" not in att:
            continue

        # Extract user information.
        FIRST_NAME = users[att]["first_name"]
        LAST_NAME = users[att]["last_name"]
        ORCID = users[att]["orcid"]
        print(f"\nChecking new cites for {FIRST_NAME} {LAST_NAME}...")

        # Query.
        if ORCID == "":
            # Query just by first name last name.
            data = query.get(
                q=f"first_author:{LAST_NAME},{FIRST_NAME}",
                fl="title,citation_count,pubdate,bibcode",
            )
        else:
            # Query also using the ORCID.
            q = (
                f"orcid_pub:{ORCID} OR orcid_user:{ORCID} OR orcid_other:{ORCID} "
                f"first_author:{LAST_NAME},{FIRST_NAME}"
            )
            data = query.get(q=q, fl="title,citation_count,pubdate,bibcode")

        # Got a bad status code?
        if data is None:
            return

        if len(data.papers) == 0:
            print(f"No paper hits for {FIRST_NAME} {LAST_NAME}")
            continue

        # To store new cites.
        new_cite_list = {}
        brand_new_paper = False

        # Path to this users' database.
        user_database = cite_tracker.get_user_database_path(att)

        # Does this user already have a database?
        if os.path.isfile(user_database):
            # Load existing database.
            database = toml.load(user_database)

            # Is there a new paper for this user that's not in the database?
            for tmp_paper in data.papers:
                if tmp_paper.bibcode not in database.keys():
                    brand_new_paper = True
                    print(
                        f"New paper for {FIRST_NAME} {LAST_NAME} ({tmp_paper.bibcode})",
                        f"adding to database...",
                    )

                    # Add new entry to database.
                    database[tmp_paper.bibcode] = {"title": tmp_paper.title, "citations": []}

            new_entry = False

            # Loop over each paper in database.
            for bibcode in database.keys():
                if verbose:
                    print(f"Checking {bibcode}...")
                new_cite_list[bibcode] = []

                # Get up-to-date cites for this paper.
                tmp_query_data = query.citations(
                    bibcode, fl="title,bibcode,author,date,doi"
                )

                # Check the query was successful
                if tmp_query_data is None:
                    print(f"Skipping {bibcode} for now, try again..")
                    continue

                # Compare to database, and see if any new cites have happened since
                # last check.
                for paper in tmp_query_data.papers:

                    if paper.bibcode not in database[bibcode]["citations"]:
                        new_cite_list[bibcode].append(paper)

                        # Record we found a new paper (to refresh the database later)
                        new_entry = True

                # Update database.
                database[bibcode]["citations"] = tmp_query_data.get_all("bibcode")

            # Report new cites.
            _print_new_cites(database, new_cite_list)

            # Update database with new entries.
            if new_entry or brand_new_paper:
                _save_database(user_database, database)
            else:
                print(f"No new cites for {FIRST_NAME} {LAST_NAME}")
        else:
            # Create new database (first time run).
            print(
                f"First time run for {FIRST_NAME} {LAST_NAME}, "
                f"creating new ADS cite tracker database file..."
            )
            _save_database(user_database, _refresh_database(data, query))



import os

import myads.cite_tracker as cite_tracker
import toml
from myads.query import ADSQueryWrapper
from tabulate import tabulate


def report():

    # Load the user database.
    assert os.path.isfile(
        cite_tracker.USER_DATABASE
    ), "No user database created yet, run 'myads --add_user'"
    users = toml.load(cite_tracker.USER_DATABASE)
    assert (
        "ads_token" in users["metadata"].keys()
    ), "No ADS API token has been added yet, run 'myads --set_ads_token <TOKEN>'"

    # Query object.
    query = ADSQueryWrapper(users["metadata"]["ads_token"])

    # Loop over each user in the database.
    for att in users.keys():
        if "user" not in att:
            continue

        # Extract this users information.
        FIRST_NAME = users[att]["first_name"]
        LAST_NAME = users[att]["last_name"]
        ORCID = users[att]["orcid"]
        print(f"\nReporting cites for {FIRST_NAME} {LAST_NAME}...")

        # Query.
        if ORCID == "":
            q = f"first_author:{LAST_NAME},{FIRST_NAME}"
        else:
            q = (
                f"orcid_pub:{ORCID} OR orcid_user:{ORCID} OR orcid_other:{ORCID} "
                f"first_author:{LAST_NAME},{FIRST_NAME}"
            )

        data = query.get(q=q, fl="title,citation_count,pubdate,bibcode")

        # Got a bad status code.
        if data is None:
            return

        # Found no papers in query.
        if len(data.papers) == 0:
            print(f"No paper hits for {FIRST_NAME} {LAST_NAME}")
            continue

        # Loop over each of my papers and print the number of cites.
        table = []
        for paper in data.papers:
            table.append(
                [
                    paper.title,
                    f"{paper.citation_count} ({paper.citations_per_year:.1f})",
                    paper.pubdate,
                    paper.link,
                ]
            )

        print(
            tabulate(
                table,
                tablefmt="grid",
                maxcolwidths=[50, None, None, None],
                headers=[
                    "Title",
                    "Citations\n(per year)",
                    "Publication\nDate",
                    "Bibcode",
                ],
            )
        )
